﻿#LoiLoNote School Sideloading Setup Version 1.0
param(
    [ValidateSet('Select', 'Install','Uninstall','ShowLog','ProvisionedInstall','ProvisionedUninstall', 'ElevatedProvisionedInstall', 'ElevatedProvisionedUninstall')]
    [string] $Action = 'Select',
    [string] $Packages,
    [switch] $Always,
    [switch] $Force
)

$ErrorActionPreference = "Stop"

Set-PSDebug -Strict

$DefaultPackageName = "LoiLoinc.51956F723A3DD.s1"
$CompanyName = "LoiLo"
$ScriptPath = (Get-Variable MyInvocation).Value.MyCommand.Path
$ScriptDirPath = Split-Path $ScriptPath -Parent
$ScriptName = $(Get-ChildItem $ScriptPath).BaseName
$WorkDirPath = Join-Path "$env:LOCALAPPDATA" (Join-Path $CompanyName $ScriptName)
$TempDirPath = Join-Path $WorkDirPath "Temp"
$LogDirPath = Join-Path $WorkDirPath "Logs"

#region Log

$LogMaxLines = 1000
$LogFileName = "log.txt"
$LogFilePath = Join-Path $LogDirPath $LogFileName
$LogBackupFileName = "log.txt.back"
$LogBackupFilePath = Join-Path $LogDirPath $LogBackupFileName

function Append-Log{

    param(
        [ValidateSet("INFO", "ERROR", ignorecase=$true)]
        [string] $type,
        [string] $msg
    )

    $now = Get-Date -Format "yyyy/MM/dd HH:mm:ss"
    $message = ("$now {0,-5} $msg" -f $type)

    if ($type -ne "ERROR") {
        Write-Host $message
    } else {
        Write-Host $message -ForegroundColor Red
    }

    if(-not (Test-Path $LogDirPath -PathType Container)){
        New-Item $LogDirPath -ItemType Directory -Force | Out-Null
    }

    if(Test-Path $LogFilePath -PathType Leaf){
        $line = (Get-Content $LogFilePath | Measure-Object –Line).Lines
        if( $line -ge $LogMaxLines ){
            if(Test-Path $LogBackupFilePath -PathType Leaf){
                Remove-Item $LogBackupFilePath -Force 
            }        
            Rename-Item $LogFilePath -NewName $LogBackupFileName 
        }
    }

    Write-Output $message | Out-File ${LogFilePath} -Append -Encoding utf8
}

#endregion

#region Logic

function Find-File {
    param( [string] $dirPath, [string] $filePattern)

    $queryPath = (Join-Path $dirPath "*.*")
    $foundItems = @(Get-ChildItem $queryPath -Include $filePattern | Where-Object { $_.Mode -NotMatch "d" })
    if($foundItems.Length -le 0) {
        return $null
    }
    return $foundItems[0]
}

function Get-TargetPackageInfo {
    param([string] $dirName)
    $packageJsonFilePath = Join-Path $dirName "package.json"
    $packageJsonContent = Get-Content -Path $packageJsonFilePath -Encoding UTF8 -Raw 
    $info = ConvertFrom-Json $packageJsonContent
    return $info
}

function Compare-Version {
    param([string] $version1, [string] $version2)

    $array1 = @($version1 -split "." , 0 , "simplematch")
    $array2 = @($version2 -split "." , 0 , "simplematch")

    for($i = 0; $i -lt 4; $i++) {
        $contains1 = ($array1.Count - 1) -ge $i;
        $contains2 = ($array2.Count - 1) -ge $i;
        if((-not $contains1) -and (-not $contains2)) {
            return 0;
        } elseif(-not $contains1) {
            return -1;
        } elseif(-not $contains2) {
            return 1;
        }

        $num1 = [int]$array1[$i]
        $num2 = [int]$array2[$i]

        if($num1 -gt $num2) {
            return 1;
        } elseif($num1 -lt $num2) {
            return -1;
        }
    }

    return 0;
}

function Get-DependencyPackages {
    param([string] $dirPath)

    $dependenciesDirPath = (Join-Path $dirPath "Dependencies")
    $dependencyPackages = @()
    if (Test-Path $dependenciesDirPath) {
        $dependencyPackages += Get-ChildItem (Join-Path $dependenciesDirPath "*.appx") | Where-Object { $_.Mode -NotMatch "d" }

        if (($Env:Processor_Architecture -eq "x86" -or $Env:Processor_Architecture -eq "amd64") -and (Test-Path (Join-Path $dependenciesDirPath "x86"))) {
            $dependencyPackages += Get-ChildItem (Join-Path $dependenciesDirPath "x86\*.appx") | Where-Object { $_.Mode -NotMatch "d" }
        } 
        if (($Env:Processor_Architecture -eq "amd64") -and (Test-Path (Join-Path $dependenciesDirPath "x64"))) {
            $dependencyPackages += Get-ChildItem (Join-Path $dependenciesDirPath "x64\*.appx") | Where-Object { $_.Mode -NotMatch "d" }
        }
        if (($Env:Processor_Architecture -eq "arm") -and (Test-Path (Join-Path $dependenciesDirPath "arm"))) {
            $dependencyPackages += Get-ChildItem (Join-Path $dependenciesDirPath "arm\*.appx") | Where-Object { $_.Mode -NotMatch "d" }
        }
    }

    return $dependencyPackages
}

function Install-Package {
    param([string] $dirPath)
    
    $bundleFilePath = Find-File $dirPath "*.appxbundle"
    if($bundleFilePath -eq $null) {
        throw "No bundle file in package."
    }

    $dependencyPackages = Get-DependencyPackages $dirPath

    if ($dependencyPackages.FullName.Count -gt 0) {
        Append-Log INFO "Found dependency($dependencyPackages)"
        Add-AppxPackage "$bundleFilePath" -DependencyPath $dependencyPackages.FullName -ForceApplicationShutdown
    } else {
        Add-AppxPackage "$bundleFilePath" -ForceApplicationShutdown
    }
}

function Install-ProvisionedPackage {
    param([string] $dirPath)
    
    $bundleFilePath = Find-File $dirPath "*.appxbundle"
    if($bundleFilePath -eq $null) {
        throw "No bundle file in package."
    }

    $dependencyPackages = Get-DependencyPackages $dirPath

    if ($dependencyPackages.FullName.Count -gt 0) {
        Append-Log INFO "Found dependency($dependencyPackages)"
        $image = Add-AppxProvisionedPackage -Online -PackagePath "$bundleFilePath" -DependencyPackagePath $dependencyPackages.FullName -SkipLicense
        Append-Log INFO ($image | Format-List | Out-String)
    } else {
        $image = Add-AppxProvisionedPackage -Online -PackagePath "$bundleFilePath" -SkipLicense
        Append-Log INFO ($image | Format-List | Out-String)
    }
}

function Uninstall-Package {
    param([string] $packageName)
    $package = Get-AppxPackage -Name $packageName
    if($package) {
        Append-Log INFO "Uninstalling..."
        Remove-AppxPackage -Package $package.PackageFullName
        Append-Log INFO "Complete!"
    } else {
        Append-Log INFO "Not installed."
    }
}

function Uninstall-ProvisionedPackage{
    param([string] $packageName)
    $packages = @(Get-AppxProvisionedPackage -Online | Where DisplayName -eq $packageName)
    if($packages) {
        foreach( $package in $packages ) {
            $packageFullname = $package.PackageName
            Append-Log INFO "Uninstalling ProvisionedPackage($packageFullname)..."
            $image = Remove-AppxProvisionedPackage -Online -PackageName $packageFullname
            Append-Log INFO ($image | Format-List | Out-String)
        }
        Append-Log INFO "Complete!" 
    } else {
        Append-Log INFO "Not installed."
    }   
}

function Run-Install {
    param([string] $packageDirPath)

    $indexFilePath = (Join-Path $packageDirPath "index.txt")
    if(-not(Test-Path $indexFilePath)) {
        throw "Not exists $indexFilePath ."
    }

    $indexFileContent = (Get-Content $indexFilePath) -as [string[]]
    if($indexFileContent.Length -le 0) {
        throw "No entry in $indexFilePath"
    }

    $targetDirName = $indexFileContent[0]
    $targetDirPath = Join-Path $packageDirPath $targetDirName
    if(-not(Test-Path $targetDirPath)) {
        throw "Not exists $targetDirPath ."
    }

    $targetPackageInfo = Get-TargetPackageInfo $targetDirPath
    $packageName = $targetPackageInfo.Name
    $targetVersion = $targetPackageInfo.Version
    $targetPublisherId = $targetPackageInfo.PublisherId

    $installedPackage = Get-AppxPackage -Name $packageName

    $removeRequired = $false
    if($installedPackage) {
        $compareResult = Compare-Version $installedPackage.Version $targetVersion
        if($compareResult -lt 0) {
            Append-Log INFO ("Currently installed $packageName Version=" + $installedPackage.Version + "(Older).")

            $installedPackagePublisherId = $installedPackage.PublisherId
            if($installedPackagePublisherId -ne $targetPublisherId) {
                Append-Log INFO "PublisherId changed.($installedPackagePublisherId to $targetPublisherId)"
                $removeRequired = $true
            }
        } elseif($compareResult -gt 0) {
            Append-Log INFO ("Currently installed $packageName Version=" + $installedPackage.Version + "(Newer).")
            if(-not $Always) {
                return
            }
            $removeRequired = $true
        } else {
            Append-Log INFO ("Already installed $packageName Version=" + $installedPackage.Version + ".")
            return
        }               
    } else {
        Append-Log INFO "Currently not installed $packageName."
    }

    $copyDirPath = Join-Path $TempDirPath $targetDirName
    Copy-Item -Path $targetDirPath -Destination $copyDirPath -Force -Recurse | Out-Null

    if( $removeRequired ) {
        Append-Log INFO "Removing Installed Version..."
        Uninstall-Package $packageName
    }

    Append-Log INFO "Installing Version=$targetVersion ..."
    Install-Package $copyDirPath
    Append-Log INFO "Complete!"
}

function Run-ProvisionedInstall {
    param([string] $packageDirPath)

    $indexFilePath = (Join-Path $packageDirPath "index.txt")
    if(-not(Test-Path $indexFilePath)) {
        throw "Not exists $indexFilePath ."
    }

    $indexFileContent = (Get-Content $indexFilePath) -as [string[]]
    if($indexFileContent.Length -le 0) {
        throw "No entry in $indexFilePath"
    }

    $targetDirName = $indexFileContent[0]
    $targetDirPath = Join-Path $packageDirPath $targetDirName
    if(-not(Test-Path $targetDirPath)) {
        throw "Not exists $targetDirPath ."
    }

    $targetPackageInfo = Get-TargetPackageInfo $targetDirPath
    $packageName = $targetPackageInfo.Name
    Uninstall-ProvisionedPackage $packageName

    $copyDirPath = Join-Path $TempDirPath $targetDirName
    Copy-Item -Path $targetDirPath -Destination $copyDirPath -Force -Recurse | Out-Null

    Append-Log INFO "Installing ProvisionedPackage..."
    Install-ProvisionedPackage $copyDirPath
    Append-Log INFO "Complete!"
}

function Ensure-TempDir {
    Append-Log INFO "Ready $TempDirPath"
    if(Test-Path $TempDirPath)
    {
        Remove-Item $TempDirPath -Recurse -Force
    }
    New-Item $TempDirPath -ItemType Directory -Force | Out-NUll
}

function Run{

    $running = $Action;
    if($running -eq 'Select') {
        $input = [int](Read-Host "Please select option below.`n0:Exit, 1:Install, 2:Uninstall, 3:ShowLog, 4:ProvisionedInstall, 5:ProvisionedUninstall`n=>")
        if($input -eq 1) {
            $running = 'Install';
        } elseif ($input -eq 2) {
            $running = 'Uninstall';
        } elseif ($input -eq 3) {
            $running = 'ShowLog';
        } elseif ($input -eq 4) {
            $running = 'ProvisionedInstall';
        } elseif ($input -eq 5) {
            $running = 'ProvisionedUninstall';
        } else {
            Append-Log INFO "Exit."
            return $true
        }
    }

    Append-Log INFO "Action=$running"

    if($running -eq 'Install') {
        Ensure-TempDir
        $packageDirPath = Join-Path $ScriptDirPath "Packages"
        if( $Packages -ne "" ){
            $packageDirPath = $Packages
        }
        Append-Log INFO "Param( Always=$Always, Packages=$packageDirPath)"
        Run-Install $packageDirPath

        return $false
    } elseif ($running -eq 'Uninstall') {
        Uninstall-Package $DefaultPackageName

        return $false
    } elseif ($running -eq 'ShowLog') {
        Start-Process "$LogDirPath" | Out-Null

        return $true
    } elseif ($running -eq 'ProvisionedInstall') {

        $relaunchArgs = '-ExecutionPolicy Unrestricted -file "' + $ScriptPath + '" -Action ElevatedProvisionedInstall'

        if($Force) {
            $relaunchArgs += ' -Force'
        }

        if( $Packages -ne "" ) {
            $relaunchArgs += ' -Packages "' + $Packages + '"'
        }
        
        Start-Process "$PsHome\PowerShell.exe" -Verb RunAs -ArgumentList $relaunchArgs -PassThru | Out-Null

        return $true
    } elseif($running -eq 'ProvisionedUninstall') {
        $relaunchArgs = '-ExecutionPolicy Unrestricted -file "' + $ScriptPath + '" -Action ElevatedProvisionedUninstall'
        Start-Process "$PsHome\PowerShell.exe" -Verb RunAs -ArgumentList $relaunchArgs -PassThru | Out-Null

        return $true
    } elseif($running -eq "ElevatedProvisionedInstall") {
        Ensure-TempDir
        $packageDirPath = Join-Path $ScriptDirPath "Packages"
        if( $Packages -ne "" ){
            $packageDirPath = $Packages
        }
        Append-Log INFO "Param( Always=$Always, Packages=$packageDirPath)"
        Run-ProvisionedInstall $packageDirPath

        return $false
    } elseif ($running -eq "ElevatedProvisionedUninstall"){
        Uninstall-ProvisionedPackage $DefaultPackageName

        return $false
    }

    return $false
}

#endregion


#
# Main script entry point
#

Append-Log INFO "[Start] $ScriptName"

$SilentExit = $false
try {
    $SilentExit = Run
} catch {
    Append-Log ERROR $_.Exception.Message
} finally {
    Append-Log INFO "[End] $ScriptName"
    if ((-not $SilentExit) -and (-not $Force)) {
        Pause
    }
}

# SIG # Begin signature block
# MIIagQYJKoZIhvcNAQcCoIIacjCCGm4CAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUYiOUgMzwviziuMgvNcmglvgA
# CLagghVvMIIElDCCA3ygAwIBAgIRAJ/qyBGw8WJHpfwg2AUjrOYwDQYJKoZIhvcN
# AQEFBQAwgZUxCzAJBgNVBAYTAlVTMQswCQYDVQQIEwJVVDEXMBUGA1UEBxMOU2Fs
# dCBMYWtlIENpdHkxHjAcBgNVBAoTFVRoZSBVU0VSVFJVU1QgTmV0d29yazEhMB8G
# A1UECxMYaHR0cDovL3d3dy51c2VydHJ1c3QuY29tMR0wGwYDVQQDExRVVE4tVVNF
# UkZpcnN0LU9iamVjdDAeFw0xNTA1MDUwMDAwMDBaFw0xNTEyMzEyMzU5NTlaMH4x
# CzAJBgNVBAYTAkdCMRswGQYDVQQIExJHcmVhdGVyIE1hbmNoZXN0ZXIxEDAOBgNV
# BAcTB1NhbGZvcmQxGjAYBgNVBAoTEUNPTU9ETyBDQSBMaW1pdGVkMSQwIgYDVQQD
# ExtDT01PRE8gVGltZSBTdGFtcGluZyBTaWduZXIwggEiMA0GCSqGSIb3DQEBAQUA
# A4IBDwAwggEKAoIBAQC8NaA2cCKBEcOyg7nTKMY2zSVrqXuyHPabUZzvNfTtCI5e
# OAj4dzwKQuDzcNyj18r1TAvP/yKcCn5o1gmiKoR7pp20qcEz4u8fF0jKOs1G5sWq
# d73jd5r6R1NAKFlDk/GkgervgLVPpwjOum68ynYMl2RZhiS7PYKQqFWxktOgpwWs
# n1MlCBBHmc2Y3mjltFB4o68BzFlDWOR2bn6sx+KeH0+wRy3IDKNJJ4B1jLsGkWUP
# kJv0utGByFxq7BTpJQm/Ixb0lUZAQCG7g5b9hh96yA0QjqL4GQdYf5+9NwJg8qTp
# nUQ/MAXkp3CZUZroF/FVyrJhiWVGp2ryWEZ+qqAHAgMBAAGjgfQwgfEwHwYDVR0j
# BBgwFoAU2u1kdBScFDyr3ZmpvVsoTYs8ydgwHQYDVR0OBBYEFC4tsApEStOHwAIH
# zpd9UGIg/Q+DMA4GA1UdDwEB/wQEAwIGwDAMBgNVHRMBAf8EAjAAMBYGA1UdJQEB
# /wQMMAoGCCsGAQUFBwMIMEIGA1UdHwQ7MDkwN6A1oDOGMWh0dHA6Ly9jcmwudXNl
# cnRydXN0LmNvbS9VVE4tVVNFUkZpcnN0LU9iamVjdC5jcmwwNQYIKwYBBQUHAQEE
# KTAnMCUGCCsGAQUFBzABhhlodHRwOi8vb2NzcC51c2VydHJ1c3QuY29tMA0GCSqG
# SIb3DQEBBQUAA4IBAQANu61gERu18A3M5kg6ej4OM9wcuerWIP6jTdDMdk7oGNh5
# 39NPmkJkI4opcoo6bGamPDoXqHBFZcZzw9DOiVT7rGkPWLAZy4affrl+61GSv5vd
# 69Fl8CV7iHzevaXItRRRvMCBMIqFOHvmef5nVZOH/k/ojQ7t83KStcKJgG3RWeMd
# DeqxOO4DnQAZpashm3nDzMI+aH69yU1pTbRkUfuyKHTiU4nOnfqt4tvOq3t+BkR0
# /QqjybenMM1J0pJk8SKmuChFdHnpp847M/mDUJR9aMAdScdgeHo8ZCbVvvoKbeQe
# 4QlTj6nFI6zHnWFCIfAsFnFJOxCvLG8a5jHxFP1sMIIFdDCCBFygAwIBAgIQJ2bu
# VutJ846r13Ci/ITeIjANBgkqhkiG9w0BAQwFADBvMQswCQYDVQQGEwJTRTEUMBIG
# A1UEChMLQWRkVHJ1c3QgQUIxJjAkBgNVBAsTHUFkZFRydXN0IEV4dGVybmFsIFRU
# UCBOZXR3b3JrMSIwIAYDVQQDExlBZGRUcnVzdCBFeHRlcm5hbCBDQSBSb290MB4X
# DTAwMDUzMDEwNDgzOFoXDTIwMDUzMDEwNDgzOFowgYUxCzAJBgNVBAYTAkdCMRsw
# GQYDVQQIExJHcmVhdGVyIE1hbmNoZXN0ZXIxEDAOBgNVBAcTB1NhbGZvcmQxGjAY
# BgNVBAoTEUNPTU9ETyBDQSBMaW1pdGVkMSswKQYDVQQDEyJDT01PRE8gUlNBIENl
# cnRpZmljYXRpb24gQXV0aG9yaXR5MIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIIC
# CgKCAgEAkehUktIKVrGsDSTdxc9EZ3SZKzejfSNwAHG8U9/E+ioSj0t/EFa9n3By
# t2F/yUsPF6c947AEYe7/EZfH9IY+Cvo+XPmT5jR62RRr55yzhaCCenavcZDX7P0N
# +pxs+t+wgvQUfvm+xKYvT3+Zf7X8Z0NyvQwA1onrayzT7Y+YHBSrfuXjbvzYqOSS
# JNpDa2K4Vf3qwbxstovzDo2a5JtsaZn4eEgwRdWt4Q08RWD8MpZRJ7xnw8outmvq
# RsfHIKCxH2XeSAi6pE6p8oNGN4Tr6MyBSENnTnIqm1y9TBsoilwie7SrmNnu4FGD
# wwlGTm0+mfqVF9p8M1dBPI1R7Qu2XK8sYxrfV8g/vOldxJuvRZnio1oktLqpVj3P
# b6r/SVi+8Kj/9Lit6Tf7urj0Czr56ENCHonYhMsT8dm74YlguIwoVqwUHZwK53Hr
# zw7dPamWoUi9PPevtQ0iTMARgexWO/bTouJbt7IEIlKVgJNp6I5MZfGRAy1wdALq
# i2cVKWlSArvX31BqVUa/oKMoYX9w0MOiqiwhqkfOKJwGRXa/ghgntNWutMtQ5mv0
# TIZxMOmm3xaG4Nj/QN370EKIf6MzOi5cHkERgWPOGHFrK+ymircxXDpqR+DDeVnW
# IBqv8mqYqnK8V0rSS527EPywTEHl7R09XiidnMy/s1Hap0flhFMCAwEAAaOB9DCB
# 8TAfBgNVHSMEGDAWgBStvZh6NLQm9/rEJlTvA73gJMtUGjAdBgNVHQ4EFgQUu69+
# Aj36pvE8hI6t7jiY7NkyMtQwDgYDVR0PAQH/BAQDAgGGMA8GA1UdEwEB/wQFMAMB
# Af8wEQYDVR0gBAowCDAGBgRVHSAAMEQGA1UdHwQ9MDswOaA3oDWGM2h0dHA6Ly9j
# cmwudXNlcnRydXN0LmNvbS9BZGRUcnVzdEV4dGVybmFsQ0FSb290LmNybDA1Bggr
# BgEFBQcBAQQpMCcwJQYIKwYBBQUHMAGGGWh0dHA6Ly9vY3NwLnVzZXJ0cnVzdC5j
# b20wDQYJKoZIhvcNAQEMBQADggEBAGS/g/FfmoXQzbihKVcN6Fr30ek+8nYEbvFS
# cLsePP9NDXRqzIGCJdPDoCpdTPW6i6FtxFQJdcfjJw5dhHk3QBN39bSsHNA7qxcS
# 1u80GH4r6XnTq1dFDK8o+tDb5VCViLvfhVdpfZLYUspzgb8c8+a4bmYRBbMelC1/
# kZWSWfFMzqORcUx8Rww7Cxn2obFshj5cqsQugsv5B5a6SE2Q8pTIqXOi6wZ7I53e
# ovNNVZ96YUWYGGjHXkBrI/V5eu+MtWuLt29G9HvxPUsE2JOAWVrgQSQdso8VYFhH
# 2+9uRv0V9dlfmrPb2LjkQLPNlzmuhbsdjrzch5vRpu/xO28QOG8wggV3MIIEX6AD
# AgECAhBWpZFYm+Qv4q5lkeb9Une7MA0GCSqGSIb3DQEBCwUAMH0xCzAJBgNVBAYT
# AkdCMRswGQYDVQQIExJHcmVhdGVyIE1hbmNoZXN0ZXIxEDAOBgNVBAcTB1NhbGZv
# cmQxGjAYBgNVBAoTEUNPTU9ETyBDQSBMaW1pdGVkMSMwIQYDVQQDExpDT01PRE8g
# UlNBIENvZGUgU2lnbmluZyBDQTAeFw0xNTAxMjMwMDAwMDBaFw0xNzAxMjIyMzU5
# NTlaMIHFMQswCQYDVQQGEwJKUDERMA8GA1UEEQwIMjMxLTAwMDYxHjAcBgNVBAgM
# FU91dHNpZGUgVW5pdGVkIFN0YXRlczEUMBIGA1UEBwwLWW9rb2hhbWFzaGkxDDAK
# BgNVBAkMAzQwMzEjMCEGA1UECQwaTmFrYWt1IE1pbmFtaW5ha2Fkb3JpIDQgNDMx
# EjAQBgNVBAoMCUxvaUxvIGluYzESMBAGA1UECwwJTG9pTG8gaW5jMRIwEAYDVQQD
# DAlMb2lMbyBpbmMwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQDnfQjP
# PsVqDE2+yob+ctESmTVgBC8dTeVHSHn3alTxcx0sr7Dgp1YBAfp1CnmwKXRYSk3Z
# KhorUMLTXBburPuPa3VTjxu2SOnppD0htfuvA1CGisOCPXN5D1dQnJwq0AMzZ049
# nsAysIlystXbVD4sBxs7TugE+YdjhhsEX4yvA21q/uqbJe9yOaAQ2qzH+5m70zIa
# Kc3tSAGms2KgaHp3zqYqoBDODMdNfpQX3xsTKIrNZrxyJ/w5a63WeltMYE3EYdRX
# FSmBO5RCj9w8bIWyhoTPIFTSfNgO2OxYK6dWedgSKjbvyC9h52oSUSIp3Fr44FS8
# xc7aL9yPs1nehWHtAgMBAAGjggGoMIIBpDAfBgNVHSMEGDAWgBQpkWD/ik366/mm
# arjP+eZLvUnOEjAdBgNVHQ4EFgQUXGiZv0bAEqJEXf8i9887SDBXZdMwDgYDVR0P
# AQH/BAQDAgeAMAwGA1UdEwEB/wQCMAAwEwYDVR0lBAwwCgYIKwYBBQUHAwMwEQYJ
# YIZIAYb4QgEBBAQDAgQQMEYGA1UdIAQ/MD0wOwYMKwYBBAGyMQECAQMCMCswKQYI
# KwYBBQUHAgEWHWh0dHBzOi8vc2VjdXJlLmNvbW9kby5uZXQvQ1BTMEMGA1UdHwQ8
# MDowOKA2oDSGMmh0dHA6Ly9jcmwuY29tb2RvY2EuY29tL0NPTU9ET1JTQUNvZGVT
# aWduaW5nQ0EuY3JsMHQGCCsGAQUFBwEBBGgwZjA+BggrBgEFBQcwAoYyaHR0cDov
# L2NydC5jb21vZG9jYS5jb20vQ09NT0RPUlNBQ29kZVNpZ25pbmdDQS5jcnQwJAYI
# KwYBBQUHMAGGGGh0dHA6Ly9vY3NwLmNvbW9kb2NhLmNvbTAZBgNVHREEEjAQgQ5s
# b2lsb0Bsb2lsby50djANBgkqhkiG9w0BAQsFAAOCAQEAh87sJw6Rz9WkR4xH7GO7
# Ogg60qIM4unPGmvnu6LUHok3rfdu/VL/gSsCi5AkQcZoTQsiFSYUrffz8JYOm/Lm
# 0bT8SF2glQ5r8tPodIgfo/X/7Svw7pmTO8kxpEbh/oZ5dVsIHHuXdDlS140VdaZj
# 6FgeZsKGgrECG8Y8AcxqQElewUOyRHW25xRyWj1R3yTz/Thv8urU+sE/4ELEPLJc
# mzUpEexzpfI7z+ctiKJbu21zXLcV9R18wUSrsJV1KASb/KUNI95HlIeoECg7lzvq
# EMmg8C9xvxd+0q31lFd126dPv710ZlbirXAIuLtUKhrwc5uVHB5YufvrCwsJCOjt
# yTCCBeAwggPIoAMCAQICEC58h8wOk0pS/pT9HLfNNK8wDQYJKoZIhvcNAQEMBQAw
# gYUxCzAJBgNVBAYTAkdCMRswGQYDVQQIExJHcmVhdGVyIE1hbmNoZXN0ZXIxEDAO
# BgNVBAcTB1NhbGZvcmQxGjAYBgNVBAoTEUNPTU9ETyBDQSBMaW1pdGVkMSswKQYD
# VQQDEyJDT01PRE8gUlNBIENlcnRpZmljYXRpb24gQXV0aG9yaXR5MB4XDTEzMDUw
# OTAwMDAwMFoXDTI4MDUwODIzNTk1OVowfTELMAkGA1UEBhMCR0IxGzAZBgNVBAgT
# EkdyZWF0ZXIgTWFuY2hlc3RlcjEQMA4GA1UEBxMHU2FsZm9yZDEaMBgGA1UEChMR
# Q09NT0RPIENBIExpbWl0ZWQxIzAhBgNVBAMTGkNPTU9ETyBSU0EgQ29kZSBTaWdu
# aW5nIENBMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAppiQY3eRNH+K
# 0d3pZzER68we/TEds7liVz+TvFvjnx4kMhEna7xRkafPnp4ls1+BqBgPHR4gMA77
# YXuGCbPj/aJonRwsnb9y4+R1oOU1I47Jiu4aDGTH2EKhe7VSA0s6sI4jS0tj4CKU
# N3vVeZAKFBhRLOb+wRLwHD9hYQqMotz2wzCqzSgYdUjBeVoIzbuMVYz31HaQOjNG
# UHOYXPSFSmsPgN1e1r39qS/AJfX5eNeNXxDCRFU8kDwxRstwrgepCuOvwQFvkBoj
# 4l8428YIXUezg0HwLgA3FLkSqnmSUs2HD3vYYimkfjC9G7WMcrRI8uPoIfleTGJ5
# iwIGn3/VCwIDAQABo4IBUTCCAU0wHwYDVR0jBBgwFoAUu69+Aj36pvE8hI6t7jiY
# 7NkyMtQwHQYDVR0OBBYEFCmRYP+KTfrr+aZquM/55ku9Sc4SMA4GA1UdDwEB/wQE
# AwIBhjASBgNVHRMBAf8ECDAGAQH/AgEAMBMGA1UdJQQMMAoGCCsGAQUFBwMDMBEG
# A1UdIAQKMAgwBgYEVR0gADBMBgNVHR8ERTBDMEGgP6A9hjtodHRwOi8vY3JsLmNv
# bW9kb2NhLmNvbS9DT01PRE9SU0FDZXJ0aWZpY2F0aW9uQXV0aG9yaXR5LmNybDBx
# BggrBgEFBQcBAQRlMGMwOwYIKwYBBQUHMAKGL2h0dHA6Ly9jcnQuY29tb2RvY2Eu
# Y29tL0NPTU9ET1JTQUFkZFRydXN0Q0EuY3J0MCQGCCsGAQUFBzABhhhodHRwOi8v
# b2NzcC5jb21vZG9jYS5jb20wDQYJKoZIhvcNAQEMBQADggIBAAI/AjnD7vjKO4ne
# DG1NsfFOkk+vwjgsBMzFYxGrCWOvq6LXAj/MbxnDPdYaCJT/JdipiKcrEBrgm7EH
# IhpRHDrU4ekJv+YkdK8eexYxbiPvVFEtUgLidQgFTPG3UeFRAMaH9mzuEER2V2rx
# 31hrIapJ1Hw3Tr3/tnVUQBg2V2cRzU8C5P7z2vx1F9vst/dlCSNJH0NXg+p+IHdh
# yE3yu2VNqPeFRQevemknZZApQIvfezpROYyoH3B5rW1CIKLPDGwDjEzNcweU51qO
# OgS6oqF8H8tjOhWn1BUbp1JHMqn0v2RH0aofU04yMHPCb7d4gp1c/0a7ayIdiAv4
# G6o0pvyM9d1/ZYyMMVcx0DbsR6HPy4uo7xwYWMUGd8pLm1GvTAhKeo/io1Lijo7M
# JuSy2OU4wqjtxoGcNWupWGFKCpe0S0K2VZ2+medwbVn4bSoMfxlgXwyaiGwwrFIJ
# kBYb/yud29AgyonqKH4yjhnfe0gzHtdl+K7J+IMUk3Z9ZNCOzr41ff9yMU2fnr0e
# bC+ojwwGUPuMJ7N2yfTm18M04oyHIYZh/r9VdOEhdwMKaGy75Mmp5s9ZJet87EUO
# eWZo6CLNuO+YhU2WETwJitB/vCgoE/tqylSNklzNwmWYBp7OSFvUtTeTRkF8B93P
# +kPvumdh/31J4LswfVyA4+YWOUunMYIEfDCCBHgCAQEwgZEwfTELMAkGA1UEBhMC
# R0IxGzAZBgNVBAgTEkdyZWF0ZXIgTWFuY2hlc3RlcjEQMA4GA1UEBxMHU2FsZm9y
# ZDEaMBgGA1UEChMRQ09NT0RPIENBIExpbWl0ZWQxIzAhBgNVBAMTGkNPTU9ETyBS
# U0EgQ29kZSBTaWduaW5nIENBAhBWpZFYm+Qv4q5lkeb9Une7MAkGBSsOAwIaBQCg
# eDAYBgorBgEEAYI3AgEMMQowCKACgAChAoAAMBkGCSqGSIb3DQEJAzEMBgorBgEE
# AYI3AgEEMBwGCisGAQQBgjcCAQsxDjAMBgorBgEEAYI3AgEVMCMGCSqGSIb3DQEJ
# BDEWBBSm59erBM0cYNwdm9O8g636tpYyjTANBgkqhkiG9w0BAQEFAASCAQAhSnMs
# xZVK7BjwwaRXdE5RroKGFU0S4qNQfPyNqfBEs06XKkdjoIEHs8PHhliaHn2hxX9r
# RDzg41qqV2ft/m6W7s48f8SYRIFKrmAmektCLlcCWxV638HVrtJJ2u4zfu1RuXs8
# 82SqzqGPM1Kc6nDrSHDmxfIHI1qLJlU82xYK34VEb45CLM71Cmov3U2IWbRhZFp4
# JRhn06ji3H3Evj6TPua0eJ9liTZGugWV5Edxt8vG8V5icfiZHwAYzMW3azGwGr/U
# R3HnODF/mXDCRXeAo+6VFjsMFNGxLAEjcLqV5h/Ow7bdBtiZJZJLX/Ok01jGPRSn
# 6QQf0nCkjhXD+pDuoYICRTCCAkEGCSqGSIb3DQEJBjGCAjIwggIuAgEAMIGrMIGV
# MQswCQYDVQQGEwJVUzELMAkGA1UECBMCVVQxFzAVBgNVBAcTDlNhbHQgTGFrZSBD
# aXR5MR4wHAYDVQQKExVUaGUgVVNFUlRSVVNUIE5ldHdvcmsxITAfBgNVBAsTGGh0
# dHA6Ly93d3cudXNlcnRydXN0LmNvbTEdMBsGA1UEAxMUVVROLVVTRVJGaXJzdC1P
# YmplY3QCEQCf6sgRsPFiR6X8INgFI6zmMAkGBSsOAwIaBQCgXTAYBgkqhkiG9w0B
# CQMxCwYJKoZIhvcNAQcBMBwGCSqGSIb3DQEJBTEPFw0xNTA3MjQxMjUwNTRaMCMG
# CSqGSIb3DQEJBDEWBBQiwekfpa+00b/nGxqxwd7HgB7jXDANBgkqhkiG9w0BAQEF
# AASCAQC4ES6F4VIAqcPnurKj76knq4ZnP05qo1vhbS/cTvaMk4fcGJVi4slLRRkN
# RQZ6/CS8p2EaF4mgFwvamB68lu9gsG/rUE9GyO8TmWM0pFU+Ly5S1fPjbhMH/tXH
# ICCqCEe42ct53n/UM18jypThsDxvd+Y809FzCspsbxOiLrbJh+jqytzsDMylz9F+
# M0/FpFXiicnaXRjKeIaBIrZ06tupwHiAd5ehipUhbmmPto4G5grvvZkR1Rull5Wb
# 6YJeoAQygLEMQjF1Hhdsxyl3+7FjmxuBnhHo1LCYV4V6yjiHRK2tL3IA47W4MUrR
# QPPwvG1RjLrKsPIdvjfGX6OYrV0S
# SIG # End signature block
